# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.16.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

import sympy as sp
sp.init_printing()

W, A, B, h, ρ, p, v2 = sp.symbols(r'W, A, B, h, \rho, p, v^2', positive=True)
ϵ, χ, κ, vrd, vru = sp.symbols(r'\epsilon, \chi, \kappa, v_r, v^r', real=True)
cs = sp.symbols(r'c_s', positive=True)
cs_rule = {
    cs : sp.sqrt((χ + (h - 1 - ϵ) / ρ * κ) / h)
}

At = sp.Matrix( [   
    [ W, ρ * vrd * W**3, 0, 0, 0 ],    
    [ (1+ϵ+χ) * vrd * W**2, ρ * h * W**2 * B**2 * (1 + 2 * W**2 * v2), 0, 0, (ρ + κ) * W**2 * vrd ],
    [ 0, 0, ρ * h * W**2 * B**2, 0, 0],
    [ 0, 0, 0, ρ * h * W**2 * B**2, 0],
    [ (1+ϵ+χ) * W**2 - χ - W, 2 * ρ * h * W**4 * vrd - ρ * W**3 * vrd, 0, 0, (ρ+κ) * W**2 - κ ]
])
At

Ar = sp.Matrix( [   
    [ W * vru, ρ * W * (1 + v2 * W**2), 0, 0, 0 ],
    [ (1+ϵ+χ) * v2 * W**2 + χ, 2 * ρ * h * W**2 * vrd * (W**2 * v2 + 1), 0, 0, (ρ + κ) * W**2 * v2 + κ ],
    [ 0, 0, ρ * h * W**2 * vrd, 0, 0],
    [ 0, 0, 0, ρ * h * W**2 * vrd, 0],
    [ (1+ϵ+κ) * W**2 * vru - W * vru, ρ * h * W**2 * (2 * v2 * W**2 + 1) - ρ * W * (W**2 * v2 + 1), 0, 0, W**2 * vru * (ρ + κ) ]
])
Ar = A * Ar
Ar

λ = sp.symbols(r'\lambda')
D = Ar - λ * At
rules = {
    vrd : vru * B**2,
    v2 : B**2 * vru**2,
    # ϵ : h - 1 - p / ρ
}
D = D.subs(rules)
D

λ0 = A * vru
D0 = D.subs(λ, λ0).expand()
display(D0)
D0.det().factor()

λp = A / (1 - v2 * cs**2) * (vru *(1-cs**2) + cs * (1 - v2) / B)
λp = λp.subs(rules)
λp = λp.subs(vru, 0)
display(λp.simplify())
Dp = D.subs(λ, λp).subs(vru, 0)
display(Dp)

# ## Verify Font, 1993

cs = sp.symbols(r'c_s', positive=True)
W, h, ρ, p = sp.symbols(r'W, h, \rho, p', positive=True)
ϵ, χ, κ, a, b, c = sp.symbols(r'\epsilon, \chi, \kappa, a, b, c', real=True)

cartAt = sp.Matrix( [   
    [ W, ρ * a * W**3, ρ * b * W**3, ρ * c * W**3, 0 ],    
    [ (1+ϵ+χ) * a * W**2, ρ * h * W**2 * (1 + 2 * a**2 * W**2), 2 * ρ * h * a * b * W**4, 2 * ρ * h * a * c * W**4, (ρ + κ) * a * W**2 ],
    [ (1+ϵ+χ) * b * W**2, 2 * ρ * h * a * b * W**4, ρ * h * W**2 * (1 + 2 * b**2 * W**2), 2 * ρ * h * b * c * W**4, (ρ + κ) * b * W**2 ],
    [ (1+ϵ+χ) * c * W**2, 2 * ρ * h * a * c * W**4, 2 * ρ * h * b * c * W**4, ρ * h * W**2 * (1 + 2 * c**2 * W**2), (ρ + κ) * c * W**2 ],
    [ (1+ϵ+χ) * W**2 - χ - W, 2 * ρ * h * a * W**4 - ρ * a * W**3, 2 * ρ * h * b * W**4 - ρ * b * W**3, 2 * ρ * h * c * W**4 - ρ * c * W**3, (ρ+κ) * W**2 - κ ]
])
cartAt

cartAx = sp.Matrix( [
    [ W * a, ρ * W * (1 + a**2 * W**2), ρ * a * b * W**3, ρ * a * c * W**3, 0 ],
    [ (1+ϵ+χ) * a**2 * W**2 + χ, 2 * ρ * h * a * W**2 * (W**2 * a**2 + 1), 2 * ρ * h * a**2 * b * W**4, 2 * ρ * h * a**2 * c * W**4, (ρ + κ) * W**2 * a**2 + κ ],
    [ (1+ϵ+χ) * a*b * W**2, ρ * h * b * W**2 * (2 * W**2 * a**2 + 1), ρ * h * a * W**2 * (1 + 2 * b**2 * W**2), 2 * ρ * h * a * b * c * W**4, (ρ + κ) * W**2 * a * b ],
    [ (1+ϵ+χ) * a*c * W**2, ρ * h * c * W**2 * (2 * W**2 * a**2 + 1), 2 * ρ * h * a * b * c * W**4, ρ * h * a * W**2 * (1 + 2 * c**2 * W**2), (ρ + κ) * W**2 * a * c ],
    [ (1+ϵ+κ) * a * W**2 - W * a, ρ * h * W**2 * (2 * a**2 * W**2 + 1) - ρ * W * (W**2 * a**2 + 1), 2 * ρ * h * a * b * W**4 - ρ * a * b * W**3, 2 * ρ * h * a * c * W**4 - ρ * a * c * W**3, W**2 * a * (ρ + κ) ]
])
cartAx

λ = sp.symbols(r'\lambda')
D = cartAx - λ * cartAt
display(D)

simplify_rules = {
    b : 0,
    c : 0
}
λ0 = a
D0 = D.subs(λ, λ0).subs(simplify_rules)
display(D0)
D0.det()

simplify_rules = {
    #a : sp.sqrt(1 - 1/W**2),
    b : 0,
    c : 0,
    #W : 1 / sp.sqrt(1 - a**2)
}
cartv2 = a**2 + b**2 + c**2
#λp = (a * (1 - cs**2) + cs * sp.sqrt((1 - cartv2) * (1 - a**2 - (b**2 + c**2) * cs**2))) / (1 - cartv2 * cs**2)
#λp = (a * (1 - cs**2) + cs * (1 - a**2)) / (1 - cartv2 * cs**2)
λp = (a + cs) / (1 + a * cs)
D1 = D.subs(λ, λp).subs(simplify_rules)
display(D1)
#D1.det()


