using dg1d
using Plots


function test_spectral_filtering(N)
  
  element = SpectralElement(N, :LGL)

  function f(x)
    if (x <= 0)
      return 0.0
    else
      2.0 * x
    end
  end

  function df(x)
    if (x <= 0)
      return 0.0
    else
      2.0
    end
  end
  
  z = element.z
  fz = f.(z)
  dfz = df.(z)

  analytical = dfz
  unfiltered = element.D * fz
  filtered = element.VFV * element.D * fz

  Plots.gr()
  plt = Plots.plot(legend=(:topleft))
  vars = [ analytical, unfiltered, filtered ]
  labels = [ "analytical", "unfiltered", "filtered" ]
  markers = [ :none, :circle, :star ]
  for (var, label, marker) in zip(vars, labels, markers)
    plot!(plt, z, var, label=label, marker = (marker))
  end
  display(plt)

  return

end
