function (eos::AbstractEquationOfState)(
      ::Type{Enthalpy},
      ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  return 1 + eps + p / rho
end


function (eos::AbstractEquationOfState)(
      ::Type{SpeedOfSound},
      ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  cs2 = eos(SquareSpeedOfSound, Density, rho, InternalEnergy, eps)
  return sqrt(cs2)
end


function (eos::AbstractEquationOfState)(
      ::Type{SquareSpeedOfSound},
      ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  chi   = eos(Derivative{Pressure, Density}, Density, rho, InternalEnergy, eps)
  kappa = eos(Derivative{Pressure, InternalEnergy}, Density, rho, InternalEnergy, eps)
  p     = eos(Pressure, Density, rho, InternalEnergy, eps)
  h     = eos(Enthalpy, Density, rho, InternalEnergy, eps)
  return (chi + p / rho^2 * kappa) / h
end


function (eos::AbstractEquationOfState)(
      ::Type{TotalEnergy},
      ::Type{Density}, rho, ::Type{InternalEnergy}, eps)

  return rho * (1 + eps)
end
