@doc """
    DBViscosity

Diffusion Based Artificial Viscosity model, cf. Yu, Hesthaven, 2019.
"""
Base.@kwdef struct DBViscosity{T_Mesh<:Mesh1d} <: AbstractArtificialViscosity
  mesh::T_Mesh
  cmax::Float64
  cb::Float64
end


function compute_viscosity!(
    mu,
    u, v,
    av::DBViscosity)

  @unpack mesh = av
  @unpack element = mesh
  @unpack D, N = element
  @unpack cmax, cb = av
  @unpack invdetJ = get_static_variables(mesh.cache)
  K, = mesh.tree.dims
  L, = dg1d.widths(mesh)

  h = L / (K * N)

  mat_u = dg1d.vreshape(u, layout(mesh))
  mat_invdetJ = dg1d.vreshape(invdetJ, layout(mesh))
  maxabs_∇u = vec(maximum(abs, @views(D * mat_u .* mat_invdetJ), dims=1))
  mu_b = cb * (h / N)^2 * maxabs_∇u

  maxabs_v = vec(maximum(abs, v, dims=1))
  mu_max = @. cmax * h / N * maxabs_v

  @. mu = min(mu_b, mu_max)

  return true
end
