abstract type AbstractShockSensor end


# """
#     linear_ramp(s, k)
#
# Ramp a sensor value `s` ∈ [0,1]
# According to Glaubitz, Phd thesis, 2020, section 10.2.5.
# """
# function linear_ramp_function(s, k)
#   if s <= k
#     return 1
#   elseif k < s < 1
#     return (1 - s) / (1 - k)
#   else
#     return 0
#   end
# end


function normalize!(weights)
  m, M = extrema(weights)
  @assert m >= 0.0 "'weights' must be positive"
  Mm = M - m
  if isapprox(Mm, 0)
    Mm = 1.0
  end
  weights .= @. (weights - m) / Mm
  return
end


"""
    shocksensor(ramp_fn::Function, weights, m, M)

Use ramping function `ramp_fn` to reduce an indicator value to a normalized value within [0,1].
`m, M` control lower and upper limits of inidicator.
`ramp_fn` should accept a sensing value `s` ∈ [0,1] and return a value in [0,1].
"""
function shocksensor!(hrsc_weights, tci_weights, ramp_fn::Function=identity)

  hrsc_weights .= tci_weights
  normalize!(hrsc_weights)

  map!(ramp_fn, hrsc_weights, hrsc_weights)

  return
end
