#######################################################################
#                  Boundary Condition implementions                   #
#######################################################################


"""
    abstract type AbstractBC end

Supertype for boundary condition types. To be used with `BoundaryConditions` interface.
"""
abstract type AbstractBC end


"""
    DirichletBC(bdryval)

Dirichlet boundary condition, `bdryval` might be time dependent.
"""
struct DirichletBC <: AbstractBC
  bdryval::Function
end
(bc::DirichletBC)(t) = bc.bdryval(t)


struct DirichletBC2<:AbstractBC end


"""
Constructor for Dirichlet BC with constant boundary value.
"""
DirichletBC(bdry_vals::NTuple) = DirichletBC((t)->bdry_vals)


struct OutflowBC    <: AbstractBC end
struct InflowBC     <: AbstractBC end
struct PeriodicBC   <: AbstractBC end
