export Cell1d, Cell2d, Cell3d,
       Cart1d, Cart2d, Cart3d,
       Tree1d, Tree2d, Tree3d,
       is_boundarycell, connect!, place_boxes


#######################################################################
#                      Computational Coordinates                      #
#######################################################################


module Cart1d
  @enum Direction::UInt8 Xmin=1 Xmax
  @enum Axis::UInt8 X=1
  for type in [ Direction, Axis ]
    eval(:(Base.to_index(d::$type) = Int(d)))
    eval(:(Base.getindex(t::Tuple, d::$type) = t[Int64(d)]))
  end
end
module Cart2d
  @enum Direction::UInt8 Xmin=1 Xmax Ymin Ymax
  @enum Axis::UInt8 X=1 Y
  for type in [ Direction, Axis ]
    eval(:(Base.to_index(d::$type) = Int(d)))
    eval(:(Base.getindex(t::Tuple, d::$type) = t[Int64(d)]))
  end
end
module Cart3d
  @enum Direction::UInt8 Xmin=1 Xmax Ymin Ymax Zmin Zmax
  @enum Axis::UInt8 X=1 Y Z
  for type in [ Direction, Axis ]
    eval(:(Base.to_index(d::$type) = Int(d)))
    eval(:(Base.getindex(t::Tuple, d::$type) = t[Int64(d)]))
  end
end


"""
    orthogonal(ax::Cart1d.Axis)
    orthogonal(ax::Cart2d.Axis)
    orthogonal(ax::Cart3d.Axis)

Return axes orthogonal to `ax`.
"""
orthogonal(ax::Cart1d.Axis) = ()
function orthogonal(ax::Cart2d.Axis)
  ax === Cart2d.X ? (Cart2d.Y,) : (Cart2d.X,)
end
function orthogonal(ax::Cart3d.Axis)
  ax === Cart3d.X && return (Cart3d.Y, Cart3d.Z)
  ax === Cart3d.Y && return (Cart3d.X, Cart3d.Z)
  return (Cart3d.X, Cart3d.Y)
end


"""
    directions(ax::Cart1d.Axis)
    directions(ax::Cart2d.Axis)
    directions(ax::Cart3d.Axis)

Return directions along axis `ax`.
"""
directions(ax::Cart1d.Axis) = (Cart1d.Xmin, Cart1d.Xmax)
function directions(ax::Cart2d.Axis)
  ax === Cart2d.X ? (Cart2d.Xmin, Cart2d.Xmax) : (Cart2d.Ymin, Cart2d.Ymax)
end
function directions(ax::Cart3d.Axis)
  ax === Cart3d.X && return (Cart3d.Xmin, Cart3d.Xmax)
  ax === Cart3d.Y && return (Cart3d.Ymin, Cart3d.Ymax)
  return (Cart3d.Zmin, Cart3d.Zmax)
end


"""
    flip(dir::Cart1d.Direction)
    flip(dir::Cart2d.Direction)
    flip(dir::Cart3d.Direction)

Invert direction `dir`.
"""
function flip(dir::Cart1d.Direction)
  dir === Cart1d.Xmin ? Cart1d.Xmax : Cart1d.Xmin
end
function flip(dir::Cart2d.Direction)
  dir === Cart2d.Xmin && return Cart2d.Xmax
  dir === Cart2d.Xmax && return Cart2d.Xmin
  dir === Cart2d.Ymin && return Cart2d.Ymax
  Cart2d.Ymin
end
function flip(dir::Cart3d.Direction)
  dir === Cart3d.Xmin && return Cart3d.Xmax
  dir === Cart3d.Xmax && return Cart3d.Xmin
  dir === Cart3d.Ymin && return Cart3d.Ymax
  dir === Cart3d.Ymax && return Cart3d.Ymin
  dir === Cart3d.Zmin && return Cart3d.Zmax
  Cart3d.Zmin
end


"""
    Coords{N_Dim}
    Coords1d
    Coords2d
    Coords3d

Integer coordinates for positioning cells in a tree.
"""
const Coords{N_Dim} = NTuple{N_Dim,Int64}
const Coords1d = Coords{1}
const Coords2d = Coords{2}
const Coords3d = Coords{3}

Base.getindex(c::Coords1d, idx::Cart1d.Axis) = c[Int(idx)]
Base.getindex(c::Coords2d, idx::Cart2d.Axis) = c[Int(idx)]
Base.getindex(c::Coords3d, idx::Cart3d.Axis) = c[Int(idx)]


"""
    is_neighbor(c1::Coords, c2::Coords)

Check whether `c1` and `c2` are direct neighbors, e.g. `distance(c1,c2) == 1`.
"""
is_neighbor(c1::Coords{N_Dim}, c2::Coords{N_Dim}) where N_Dim = distance(c1, c2) == 1

"""
    is_neighbor(c1::Coords1d, c2::Coords1d, dir::Cart1d.Direction)
    is_neighbor(c1::Coords2d, c2::Coords2d, dir::Cart2d.Direction)
    is_neighbor(c1::Coords3d, c2::Coords3d, dir::Cart3d.Direction)

Check whether `c2` is the neighbor of `c1` in direction `dir`.
"""
function is_neighbor(c1::Coords1d, c2::Coords1d, dir::Cart1d.Direction)
  sd = signed_distance(c1, c2)
  abs(sd) != 1 && return false
  dir === Cart1d.Xmin && return sd == -1
  dir === Cart1d.Xmax && return sd == 1
end
function is_neighbor(c1::Coords2d, c2::Coords2d, dir::Cart2d.Direction)
  distance(c1, c2) != 1 && return false
  dist = c2 .- c1
  dir === Cart2d.Xmin && return dist[Cart2d.X] == -1
  dir === Cart2d.Xmax && return dist[Cart2d.X] == 1
  dir === Cart2d.Ymin && return dist[Cart2d.Y] == -1
  dir === Cart2d.Ymax && return dist[Cart2d.Y] == 1
end
function is_neighbor(c1::Coords3d, c2::Coords3d, dir::Cart3d.Direction)
  distance(c1, c2) != 1 && return false
  dist = c2 .- c1
  dir === Cart3d.Xmin && return dist[Cart3d.X] == -1
  dir === Cart3d.Xmax && return dist[Cart3d.X] == 1
  dir === Cart3d.Ymin && return dist[Cart3d.Y] == -1
  dir === Cart3d.Ymax && return dist[Cart3d.Y] == 1
  dir === Cart3d.Zmin && return dist[Cart3d.Z] == -1
  dir === Cart3d.Zmax && return dist[Cart3d.Z] == 1
end


"""
    touch_direction(c1::Coords1d, c2::Coords1d)
    touch_direction(c1::Coords2d, c2::Coords2d)
    touch_direction(c1::Coords3d, c2::Coords3d)

Return direction in which `c2` touches `c1`.
"""
function touch_direction(c1::Coords1d, c2::Coords1d)
  !is_neighbor(c1, c2) && return nothing
  is_neighbor(c1, c2, Cart1d.Xmax) ? Cart1d.Xmax : Cart1d.Xmin
end
function touch_direction(c1::Coords2d, c2::Coords2d)
  !is_neighbor(c1, c2) && return nothing
  is_neighbor(c1, c2, Cart2d.Xmin) && return Cart2d.Xmin
  is_neighbor(c1, c2, Cart2d.Xmax) && return Cart2d.Xmax
  is_neighbor(c1, c2, Cart2d.Ymin) && return Cart2d.Ymin
  is_neighbor(c1, c2, Cart2d.Ymax) && return Cart2d.Ymax
end
function touch_direction(c1::Coords3d, c2::Coords3d)
  !is_neighbor(c1, c2) && return nothing
  is_neighbor(c1, c2, Cart3d.Xmin) && return Cart3d.Xmin
  is_neighbor(c1, c2, Cart3d.Xmax) && return Cart3d.Xmax
  is_neighbor(c1, c2, Cart3d.Ymin) && return Cart3d.Ymin
  is_neighbor(c1, c2, Cart3d.Ymax) && return Cart3d.Ymax
  is_neighbor(c1, c2, Cart3d.Zmin) && return Cart3d.Zmin
  is_neighbor(c1, c2, Cart3d.Zmax) && return Cart3d.Zmax
end


"""
    distance(c1::Coords, c2::Coords)

Discrete distance functional.
"""
distance(c1::Coords, c2::Coords) = sum(abs.(c2 .- c1))
"""
    signed_distance(c1::Coords, c2::Coords)

Signed discrete distance functional.
"""
signed_distance(c1::Coords, c2::Coords) = sum(c2 .- c1)
# signed_distance(c1::Coords1d, c2::Coords1d, dir::Cart1d.Direction) = c2[dir] - c1[dir]
# signed_distance(c1::Coords2d, c2::Coords2d, dir::Cart2d.Direction) = c2[dir] - c1[dir]
# signed_distance(c1::Coords3d, c2::Coords3d, dir::Cart3d.Direction) = c2[dir] - c1[dir]


# Idea: Write macro
# @coordinates SphericalCoords3d begin
#   Axis : r θ φ
# end
#
# translates into
#
# module SphericalCoords3d
#   @enum Direction::UInt8 rmin=1 rmax θmin θmax φmin φmax
#   @enum Axis::UInt8 r=1 θ φ
#   for type in [ Direction, Axis ]
#     @eval :(Base.to_index(d::$type) = Int(d))
#   end
# end


#######################################################################
#                                Cell                                 #
#######################################################################


"""
    mutable struct Cell{N_Dim,N_Sides}
    mutable struct Cell1d
    mutable struct Cell2d
    mutable struct Cell3d

Building block for a computational mesh to encode topological info of mesh layout.
"""
Base.@kwdef struct Cell{N_Dim,N_Sides}
  index::Int64 = 0 # change to parent for AMR
  neighbors::MVector{N_Sides,Int64} = MVector{N_Sides,Int64}(0 for _ = 1:N_Sides)
  # neighbors::NTuple{N_Sides,Int64} = Tuple(zeros(Int64,  N_Sides))
  # for AMR add weights/levels here
  # hlevel
  # plevel
  # ...
end
const Cell1d = Cell{1,2}
const Cell2d = Cell{2,4}
const Cell3d = Cell{3,6}


"""
    is_boundarycell(cell::Cell)

`cells` is a boundary cell if it misses a neighbor in any direction.
"""
function is_boundarycell(cell::Cell)
  for ngbr in cell.neighbors
    iszero(ngbr) && return true
  end
  return false
end


"""
    has_neighbor(cell::Cell1d, dir::Cart1d.Direction)
    has_neighbor(cell::Cell2d, dir::Cart2d.Direction)
    has_neighbor(cell::Cell3d, dir::Cart3d.Direction)

Check if `cells` has a neighbor in direction `dir`.
"""
has_neighbor(cell::Cell1d, dir::Cart1d.Direction) = cell.neighbors[dir] != 0
has_neighbor(cell::Cell2d, dir::Cart2d.Direction) = cell.neighbors[dir] != 0
has_neighbor(cell::Cell3d, dir::Cart3d.Direction) = cell.neighbors[dir] != 0


"""
    is_neighbor(c1::Cell1d, c2::Cell1d, dir::Cart1d.Direction)
    is_neighbor(c1::Cell2d, c2::Cell2d, dir::Cart2d.Direction)
    is_neighbor(c1::Cell3d, c2::Cell3d, dir::Cart3d.Direction)

Check whether `c2` is the neighbor of `c1` in direction `dir`.
"""
is_neighbor(c1::Cell1d, c2::Cell1d, dir::Cart1d.Direction) = _is_neighbor(c1, c2, dir)
is_neighbor(c1::Cell2d, c2::Cell2d, dir::Cart2d.Direction) = _is_neighbor(c1, c2, dir)
is_neighbor(c1::Cell3d, c2::Cell3d, dir::Cart3d.Direction) = _is_neighbor(c1, c2, dir)
function _is_neighbor(cell1, cell2, dir)
  return cell1.index === cell2.neighbors[Int(dir)]
end
is_neighbor(c1::Cell1d, c2::Cell1d) = any(is_neighbor(c1, c2, dir) for dir in instances(Cart1d.Direction))
is_neighbor(c1::Cell2d, c2::Cell2d) = any(is_neighbor(c1, c2, dir) for dir in instances(Cart2d.Direction))
is_neighbor(c1::Cell3d, c2::Cell3d) = any(is_neighbor(c1, c2, dir) for dir in instances(Cart3d.Direction))


"""
    AbstractTree

Supertype for abstract representation of a regular mesh of cells.
"""
abstract type AbstractTree end


"""
    struct Tree{N_Dim,N_Sides} <: AbstractTree
    struct Tree1d
    struct Tree2d
    struct Tree3d

Abstract representation of a rectangular mesh of cells.
"""
Base.@kwdef struct Tree{N_Dim,N_Sides} <: AbstractTree
  cells::Vector{Cell{N_Dim,N_Sides}}
  bdrycells::Vector{Cell{N_Dim,N_Sides}}
  coords::Vector{NTuple{N_Dim,Int64}}
  dims::NTuple{N_Dim,Int64}
  periodic::NTuple{N_Dim,Bool}

  function Tree{N_Dim,N_Sides}(dims::Vararg{Int64,N_Dim};
                               periodic::NTuple{N_Dim,Bool} = Tuple(false for _ in 1:N_Dim)) where {N_Dim,N_Sides}
    @assert all(dims .> 0)
    coords = coordinates_snake_curve(dims...)
    ncells = prod(dims)
    cells = [ Cell{N_Dim,2*N_Dim}(index=i) for i in 1:ncells ]
    bdrycells = Cell{N_Dim,2*N_Dim}[]
    resize!(cells, prod(dims))
    tree = new{N_Dim,2*N_Dim}(cells, bdrycells, coords, dims, periodic)
    connect!(tree)
    for cell in cells
      is_boundarycell(cell) && push!(tree.bdrycells, cell)
    end
    return tree
  end
end
const Tree1d = Tree{1,2}
const Tree2d = Tree{2,4}
const Tree3d = Tree{3,6}


# forward iterator interface to cells
Base.iterate(t::Tree) = iterate(t.cells)
Base.iterate(t::Tree, idx::Int) = iterate(t.cells, idx)
Base.length(t::Tree) = n_cells(t)
Base.eltype(t::Tree) = eltype(t.cells)
Base.getindex(t::Tree, idx) = t.cells[idx]


"""
    n_cells(t::Tree)

Number of cells contained in `t`.
"""
n_cells(t::Tree) = length(t.cells)


"""
    connet!(t::Tree)

Update cell neighbors according to current cell distribution.
"""
function connect!(t::Tree)
  _connect_bulk!(t)
  _connect_periodic!(t)
end


function _connect_bulk!(t::Tree{N_Dim}) where N_Dim

  N_Sides = 2*N_Dim
  ncells = n_cells(t)

  # connect closest cells with one another
  closest = Tuple{Int64,Coords{N_Dim}}[]
  for i = 1:ncells

    cell = t.cells[i]
    coord = t.coords[i]
    empty!(closest)
    sizehint!(closest, N_Sides)
    for j = 1:ncells
      coord_j = t.coords[j]
      dist = distance(coord, coord_j)
      dist == 1 && push!(closest, (j,coord_j))
      length(closest) == N_Sides && break
    end

    for (j,j_coord) in closest
      dir = touch_direction(coord, j_coord)
      cell.neighbors[dir] = j
    end

  end

end


function _connect_periodic!(t::Tree1d)

  px, = t.periodic
  nx, = t.dims
  if px
    t.cells[1].neighbors[Cart1d.Xmin] = nx
    t.cells[nx].neighbors[Cart1d.Xmax] = 1
  end

end


function _connect_periodic!(t::Tree2d)
  px, py = t.periodic
  px && _connect_periodic!(t, Cart2d.X)
  py && _connect_periodic!(t, Cart2d.Y)
end


function _connect_periodic!(t::Tree3d)
  px, py, pz = t.periodic
  px && _connect_periodic!(t, Cart3d.X)
  py && _connect_periodic!(t, Cart3d.Y)
  pz && _connect_periodic!(t, Cart3d.Z)
end


function _connect_periodic!(t::Tree2d, ax::Cart2d.Axis)

  dirmin, dirmax = directions(ax)
  ortho_ax, = orthogonal(ax)

  # filter boundary cells, returns a list of indices for the t.cells array
  i_bdrycells = findall(t.cells) do cell
    cell.neighbors[dirmin] == 0 || cell.neighbors[dirmax] == 0
  end
  length(i_bdrycells) == 0 && error("Could not find any unconnected boundaries")
  # determine those is for which a dirmin neighbor is missing
  is_of_i_bdrycells_dirmin = findall(i_bdrycells) do i
    cell = t.cells[i]
    cell.neighbors[dirmin] == 0
  end
  i_bdrycells_dirmin = i_bdrycells[is_of_i_bdrycells_dirmin]
  # demand that the number of boundary cells connected between the periodic sides is the same
  2*length(i_bdrycells_dirmin) != length(i_bdrycells) &&
    error("Encountered an unbalanced number of boundaries to connect")
  i_bdrycells_dirmax = setdiff(i_bdrycells, i_bdrycells_dirmin)

  # connect pairs of cells with same coordinates in orthogonal directions
  # only iterating the ones which miss a dirmin neighbor
  for i_dirmin in i_bdrycells_dirmin
    cell = t.cells[i_dirmin]
    coords = t.coords[i_dirmin]
    j = findfirst(i_bdrycells_dirmax) do jj
      coords[ortho_ax] == t.coords[jj][ortho_ax] && return true
      return false
    end
    isnothing(j) && error("Could not find opposing cell for '$cell'")
    i_dirmax = i_bdrycells_dirmax[j]
    cell.neighbors[dirmin]              = i_dirmax
    t.cells[i_dirmax].neighbors[dirmax] = i_dirmin
  end

end


function _connect_periodic!(t::Tree3d, ax::Cart3d.Axis)

  dirmin, dirmax = directions(ax)
  ortho_ax1, ortho_ax2 = orthogonal(ax)

  # filter boundary cells, returns a list of indices for the t.cells array
  i_bdrycells = findall(t.cells) do cell
    cell.neighbors[dirmin] == 0 || cell.neighbors[dirmax] == 0
  end
  length(i_bdrycells) == 0 && error("Could not find any unconnected boundaries")
  # determine those is for which a dirmin neighbor is missing
  is_of_i_bdrycells_dirmin = findall(i_bdrycells) do i
    cell = t.cells[i]
    cell.neighbors[dirmin] == 0
  end
  i_bdrycells_dirmin = i_bdrycells[is_of_i_bdrycells_dirmin]
  # demand that the number of boundary cells connected between the periodic sides is the same
  2*length(i_bdrycells_dirmin) != length(i_bdrycells) &&
    error("Encountered an unbalanced number of boundaries to connect")
  i_bdrycells_dirmax = setdiff(i_bdrycells, i_bdrycells_dirmin)


  # connect pairs of cells with same coordinates in orthogonal directions
  # only iterating the ones which miss a dirmin neighbor
  for i_dirmin in i_bdrycells_dirmin
    cell = t.cells[i_dirmin]
    coords = t.coords[i_dirmin]
    j = findfirst(i_bdrycells_dirmax) do jj
      coords[ortho_ax1] == t.coords[jj][ortho_ax1] &&
      coords[ortho_ax2] == t.coords[jj][ortho_ax2] && return true
      return false
      coords[ortho_ax] == t.coords[jj][ortho_ax] && return true
      return false
    end
    isnothing(j) && error("Could not find opposing cell for '$cell'")
    i_dirmax = i_bdrycells_dirmax[j]
    cell.neighbors[dirmin]              = i_dirmax
    t.cells[i_dirmax].neighbors[dirmax] = i_dirmin
  end

end


"""
    coordinates_snake_curve(Nx)
    coordinates_snake_curve(Nx,Ny)
    coordinates_snake_curve(Nx,Ny,Nz)

Generate coordinates of a snake curve that fills a 1d/2d/3d space with dimensions
`(Nx)/(Nx,Ny)/(Nx,Ny,Nz)`.
"""
function coordinates_snake_curve(Nx)
  coords = coordinates_snake_curve(Nx, 1, 1)
  coords = [ tuple(c[1]) for c in coords ]
  return coords
end
function coordinates_snake_curve(Nx, Ny)
  coords = coordinates_snake_curve(Nx, Ny, 1)
  coords = [ c[1:2] for c in coords ]
  return coords
end
function coordinates_snake_curve(Nx, Ny, Nz)
  NNN = Nx*Ny*Nz
  ix, iy, iz = 1, 1, 1
  ix_dir, iy_dir = 1, 1
  ijks = NTuple{3,Int64}[]
  for _ = 1:NNN
    push!(ijks, (ix,iy,iz))
    if ix_dir > 0
      if ix < Nx
        ix += 1
      else
        if iy_dir > 0
          if iy < Ny
            iy += 1
          else
            iz += 1
            iy_dir *= -1
          end
        else
          if iy > 1
            iy -= 1
          else
            iz += 1
            iy_dir *= -1
          end
        end
        ix_dir *= -1
      end
    else
      if ix > 1
        ix -= 1
      else
        if iy_dir > 0
          if iy < Ny
            iy += 1
          else
            iz += 1
            iy_dir *= -1
          end
        else
          if iy > 1
            iy -= 1
          else
            iz += 1
            iy_dir *= -1
          end
        end
        ix_dir *= -1
      end
    end
  end
  return ijks
end


function error_failed_placement(cell, prev_cell)
  error("Failed to place box at '$cell' because it is not a neighbor of '$prev_cell'")
end


"""
    place_boxes(t::Tree1d, x_extends)
    place_boxes(t::Tree2d, x_extends, y_extends)
    place_boxes(t::Tree3d, x_extends, y_extends, z_extends)

Place boxes centered onto the nodes of tree `t`.
`x_extends, y_extends, z_extends` provide the extrema of the rectangular space all
boxes will fill.
"""
function place_boxes(t::Tree1d, x_extends)

  boxes = Box1d[]
  sizehint!(boxes, n_cells(t))
  nx, = t.dims
  xmin, xmax = x_extends
  # use a range, because it handles equal paritioning of intervals accurately
  # in particular, the endpoints will be preserved
  xrng = range(xmin, xmax, length=nx+1)

  # first box
  push!(boxes, Box1d( (xrng[1],xrng[2]) ))
  for i in 2:n_cells(t)

    prev_cell, cell = t.cells[i-1], t.cells[i]
    !is_neighbor(prev_cell, cell) && error_failed_placement(cell, prev_cell)

    prev_coords, coords = t.coords[i-1], t.coords[i]
    touch_at = touch_direction(prev_coords, coords)
    ix, = prev_coords
    @assert touch_at === Cart1d.Xmax
    ixmin, ixmax = ix+1, ix+2

    push!(boxes, Box1d( (xrng[ixmin], xrng[ixmax]) ))
  end

  return boxes
end


function place_boxes(t::Tree2d, x_extends, y_extends)

  boxes = Box2d[]
  sizehint!(boxes, n_cells(t))
  nx, ny = t.dims
  xmin, xmax = x_extends
  ymin, ymax = y_extends
  # use a range, because it handles equal paritioning of intervals accurately
  # in particular, the endpoints will be preserved
  xrng = range(xmin, xmax, length=nx+1)
  yrng = range(ymin, ymax, length=ny+1)

  # first box
  push!(boxes, Box2d( (xrng[1],xrng[2]), (yrng[1],yrng[2]) ))
  for i in 2:n_cells(t)

    prev_cell, cell = t.cells[i-1], t.cells[i]
    !is_neighbor(prev_cell, cell) && error_failed_placement(cell, prev_cell)

    prev_coords, coords = t.coords[i-1], t.coords[i]
    touch_at = touch_direction(prev_coords, coords)
    prev_box = last(boxes)
    ix, iy = prev_coords
    ixmin, ixmax, iymin, iymax = if touch_at === Cart2d.Xmax
      ix+1, ix+2, iy, iy+1
    elseif touch_at === Cart2d.Xmin
      ix-1, ix, iy, iy+1
    elseif touch_at === Cart2d.Ymax
      ix, ix+1, iy+1, iy+2
    else
      ix, ix+1, iy-1, iy
    end

    push!(boxes, Box2d( (xrng[ixmin],xrng[ixmax]),
                        (yrng[iymin],yrng[iymax]) ))
  end

  return boxes
end


function place_boxes(t::Tree3d, x_extends, y_extends, z_extends)

  boxes = Box3d[]
  sizehint!(boxes, n_cells(t))
  nx, ny, nz = t.dims
  xmin, xmax = x_extends
  ymin, ymax = y_extends
  zmin, zmax = y_extends
  # use a range, because it handles equal paritioning of intervals accurately
  # in particular, the endpoints will be preserved
  xrng = range(xmin, xmax, length=nx+1)
  yrng = range(ymin, ymax, length=ny+1)
  zrng = range(zmin, zmax, length=nz+1)

  # first box
  # push!(boxes, Box3d( (xmin,xmin+dx), (ymin,ymin+dy), (zmin,zmin+dz) ))
  push!(boxes, Box3d( (xrng[1],xrng[2]), (yrng[1],yrng[2]), (zrng[1],zrng[2]) ))
  for i in 2:n_cells(t)

    prev_cell, cell = t.cells[i-1], t.cells[i]
    !is_neighbor(prev_cell, cell) && error_failed_placement(cell, prev_cell)

    prev_coords, coords = t.coords[i-1], t.coords[i]
    touch_at = touch_direction(prev_coords, coords)
    ix, iy, iz = prev_coords
    ixmin, ixmax, iymin, iymax, izmin, izmax = if touch_at === Cart3d.Xmax
      ix+1, ix+2, iy, iy+1, iz, iz+1
    elseif touch_at === Cart3d.Xmin
      ix-1, ix, iy, iy+1, iz, iz+1
    elseif touch_at === Cart3d.Ymax
      ix, ix+1, iy+1, iy+2, iz, iz+1
    elseif touch_at === Cart3d.Ymin
      ix, ix+1, iy-1, iy, iz, iz+1
    elseif touch_at === Cart3d.Zmax
      ix, ix+1, iy, iy+1, iz+1, iz+2
    else
      ix, ix+1, iy, iy+1, iz-1, iz
    end

    push!(boxes, Box3d( (xrng[ixmin],xrng[ixmax]),
                        (yrng[iymin],yrng[iymax]),
                        (zrng[izmin],zrng[izmax]) ))
  end

  return boxes
end


function Base.show(io::IO, t::Tree)
  print("""
        $(typeof(t))
          # cells = $(length(t.cells))
        """)
end



#######################################################################
#                              AMR stuff                              #
#######################################################################


# function update!(t::Tree)
#   TODO(update!)
# end

# const Tile1d = MVector{2,Int64}
# const Tile2d = MVector{4,Int64}
# is_boundary_cell(tcid::Tile{Cell}) = any(Tuple(cid.type == BoundaryCell for cid in tcid.cellids))
# is_volume_cell(tcid::Tile{Cell})   = any(Tuple(cid.type == VolumeCell for cid in tcid.cellids))
# is_no_cell(tcid::Tile{Cell})       = any(Tuple(cid.type == NoCell for cid in tcid.cellids))
# Base.length(tcid::Tile1d) = 2
# Base.length(tcid::Tile2d) = 4


# const TileNeighbours1d = Neighbours1d
# const TileNeighbours2d = MVector{4,Union{Int64,Tile1d}}
# const TileNeighbours3d = MVector{6,Union{Int64,Tile2d}}


# # h-refinement
# has_tiles(nghbr::Neighbours1d) = false
# has_tilled(nghbr::Neighbours) = any( #= member isa Tile for memeber in nghbr =#)
# find_tiles(nghbr::Neighbours1d) = nothing
# findtilled(nghbr::Neighbours) = [ (side, id) for (side, id) in nghbr ]


# mutable struct Cell{N_Dim,T_Neighbours}
#   index::Int64
#   neighbors::T_Neighbours
#   # level::Int64
#   # parent::Int64
#   # children::Vector{Int64}
#   # hlevel::Int64
#   # plevel::Int64
#   # tlevel::Int64
# end
