@testset "TensorBasis" begin

  nx, ny, nz = 7,8,9
  elx, ely, elz = [ SpectralElement(Np-1, :LGL) for Np in (nx,ny,nz) ]
  tb1 = TensorBasis1d(elx)
  tb2 = TensorBasis2d(elx, ely)
  tb3 = TensorBasis3d(elx, ely, elz)

  println(tb1)
  println(tb2)
  println(tb3)

  @test ndims(tb1) == 1
  @test ndims(tb2) == 2
  @test ndims(tb3) == 3

  @test size(tb1) == (nx,)
  @test size(tb2) == (nx,ny,)
  @test size(tb3) == (nx,ny,nz)

  @test n_points(tb1) == nx
  @test n_points(tb2) == nx*ny
  @test n_points(tb3) == nx*ny*nz

  @test n_points_boundary(tb1) == 2
  @test n_points_boundary(tb2) == 2*ny+2*nx - 4
  @test n_points_boundary(tb3) == 2*ny*nz+2*nx*nz+2*nx*ny - 16

  @test n_points_per_boundary(tb1) == (1,)
  @test n_points_per_boundary(tb2) == (ny,nx)
  @test n_points_per_boundary(tb3) == (ny*nz,nx*nz,nx*ny)

  m1d = collect(Float64, 1:n_points(tb1))
  m2d = collect(Float64, 1:n_points(tb2))
  m3d = collect(Float64, 1:n_points(tb3))
  m2d = reshape(m2d, size(tb2))
  m3d = reshape(m3d, size(tb3))

  m1d[[1,end]]      .= NaN

  m2d[1:end,1]      .= NaN
  m2d[1:end,end]    .= NaN
  m2d[1,1:end]      .= NaN
  m2d[end,1:end]    .= NaN

  m3d[1:end,1:end,1]  .= NaN
  m3d[1:end,1:end,end].= NaN
  m3d[1:end,1,1:end]  .= NaN
  m3d[1:end,end,1:end].= NaN
  m3d[1,1:end,1:end]  .= NaN
  m3d[end,1:end,1:end].= NaN

  volrng1d = range_volume(tb1)
  volrng2d = range_volume(tb2)
  volrng3d = range_volume(tb3)
  @test length(volrng1d) == n_points(tb1)
  @test length(volrng2d) == n_points(tb2)
  @test length(volrng3d) == n_points(tb3)
  # strides_bdry_1d = strides_boundary(tb1)
  # strides_bdry_2d = strides_boundary(tb2)
  # strides_bdry_3d = strides_boundary(tb3)

  # println(strides_bdry_3d)
  # display(m3d[1,:,:])
  # display(m3d[1:30])

  # for (tb,m) in [ (tb1,m1d), (tb2,m2d), (tb3,m3d) ]
  #   strides = strides_boundary(tb)
  #   for (i,s) in enumerate(strides)
  #     i > 1 && break
  #     @test all(isnan.(m[s])) == true
  #   end
  # end
  # @test all(isnan.(m1d[bdryrng1d])) == true
  # @test all(isnan.(m2d[bdryrng2d])) == true
  # @test all(isnan.(m3d[bdryrng3d])) == true

end
